"""Public proxy interface with SNI."""
import asyncio
import ipaddress
import logging
from contextlib import suppress
from typing import Optional

import async_timeout

from ..exceptions import (MultiplexerTransportClose, MultiplexerTransportError,
                          ParseSNIError)
from ..multiplexer.core import Multiplexer
from .peer_manager import PeerManager
from .sni import parse_tls_sni

_LOGGER = logging.getLogger(__name__)


class SNIProxy:
    """SNI Proxy class."""

    def __init__(self, peer_manager: PeerManager, host=None, port=None):
        """Initialize SNI Proxy interface."""
        self._peer_manager = peer_manager
        self._loop = asyncio.get_event_loop()
        self._host = host
        self._port = port or 443
        self._server = None

    async def start(self):
        """Start Proxy server."""
        self._server = await asyncio.start_server(
            self.handle_connection, host=self._host, port=self._port)

    async def stop(self):
        """Stop proxy server."""
        self._server.close()
        await self._server.wait_closed()

    async def handle_connection(self,
                                reader: asyncio.StreamReader,
                                writer: asyncio.StreamWriter,
                                data: Optional[bytes] = None):
        """Internal handler for incoming requests."""
        if not data:
            try:
                async with async_timeout.timeout(2):
                    client_hello = await reader.read(1024)
            except asyncio.TimeoutError:
                _LOGGER.warning("Close SNI handshake because timeout")
                writer.close()
                return
        else:
            client_hello = data

        # Connection closed before data received
        if not client_hello:
            with suppress(OSError):
                writer.close()
            return

        try:
            # Read Hostname
            try:
                hostname = parse_tls_sni(client_hello)
            except ParseSNIError:
                _LOGGER.warning(
                    "Receive invalid ClientHello on public Interface")
                return

            # Peer available?
            if not self._peer_manager.peer_available(hostname):
                _LOGGER.warning("Hostname %s not connected", hostname)
                return
            peer = self._peer_manager.get_peer(hostname)

            # Proxy data over mutliplexer to client
            _LOGGER.debug("Processing for hostname % started", hostname)
            await self._proxy_peer(peer.multiplexer, client_hello, reader,
                                   writer)

        finally:
            with suppress(OSError):
                writer.close()

    async def _proxy_peer(self, multiplexer: Multiplexer, client_hello: bytes,
                          reader: asyncio.StreamReader,
                          writer: asyncio.StreamWriter):
        """Proxy data between end points."""
        transport = writer.transport
        ip_address = ipaddress.ip_address(writer.get_extra_info("peername")[0])

        # Open multiplexer channel
        try:
            channel = await multiplexer.create_channel(ip_address)
            await channel.write(client_hello)
        except MultiplexerTransportError:
            _LOGGER.error("Transport to peer fails")
            return

        from_proxy = None
        from_peer = None
        try:
            # Process stream into multiplexer
            while not transport.is_closing():
                if not from_proxy:
                    from_proxy = self._loop.create_task(reader.read(4096))
                if not from_peer:
                    from_peer = self._loop.create_task(channel.read())

                # Wait until data need to be processed
                await asyncio.wait([from_proxy, from_peer],
                                   return_when=asyncio.FIRST_COMPLETED)

                # From proxy
                if from_proxy.done():
                    if from_proxy.exception():
                        raise from_proxy.exception()

                    await channel.write(from_proxy.result())
                    from_proxy = None

                # From peer
                if from_peer.done():
                    if from_peer.exception():
                        raise from_peer.exception()

                    writer.write(from_peer.result())
                    from_peer = None

        except MultiplexerTransportError:
            _LOGGER.debug("Transport closed by Proxy for %s", channel.uuid)
            await multiplexer.delete_channel(channel)

        except MultiplexerTransportClose:
            _LOGGER.debug("Peer close connection for %s", channel.uuid)

        finally:
            if from_peer and not from_peer.done():
                from_peer.cancel()
            if from_proxy and not from_proxy.done():
                from_proxy.cancel()



"""Helper for handle aiohttp internal server."""
from contextlib import suppress
import socket
import ssl
import logging

from aiohttp.web import AppRunner, SockSite

from ..client.client_peer import ClientPeer
from ..client.connector import Connector

_LOGGER = logging.getLogger(__name__)


class SniTunClientAioHttp:
    """Help to handle a internal aiohttp app runner."""

    def __init__(self,
                 runner: AppRunner,
                 context: ssl.SSLContext,
                 snitun_server: str,
                 snitun_port=None):
        """Initialize SniTunClient with aiohttp."""
        self._connector = None
        self._client = ClientPeer(snitun_server, snitun_port)
        self._socket = socket.socket()
        self._server_name = "{}:{}".format(snitun_server, snitun_port)

        # Init interface
        self._socket.setblocking(False)
        self._socket.bind(("127.0.0.1", 0))
        self._site = SockSite(runner, self._socket, ssl_context=context)

    @property
    def is_connected(self) -> bool:
        """Return True if we are connected to snitun."""
        return self._client.is_connected

    @property
    def whitelist(self):
        """Return whitelist from connector."""
        if self._connector:
            return self._connector.whitelist
        return set()

    async def start(self, whitelist=False):
        """Start internal server."""
        await self._site.start()

        host, port = self._socket.getsockname()[:2]
        self._connector = Connector(host, port, whitelist)

        _LOGGER.info("AioHTTP snitun client started on %s:%s", host, port)

    async def stop(self):
        """Stop internal server."""
        await self.disconnect()
        with suppress(OSError):
            self._socket.close()

        # pylint: disable=protected-access
        self._site._runner._unreg_site(self._site)
        _LOGGER.info("AioHTTP snitun client closed")

    async def connect(self, fernet_key, aes_key, aes_iv):
        """Connect to SniTun server."""
        if self._client.is_connected:
            return
        await self._client.start(self._connector, fernet_key, aes_key, aes_iv)
        _LOGGER.info("AioHTTP snitun client connected to: %s",
                     self._server_name)

    async def disconnect(self):
        """Disconnect from SniTun server."""
        if not self._client.is_connected:
            return
        await self._client.stop()
        _LOGGER.info("AioHTTP snitun client disconnected from: %s",
                     self._server_name)



"""Manage peer connections."""
from datetime import datetime
import json
import logging
from typing import List

from cryptography.fernet import Fernet, InvalidToken, MultiFernet

from ..exceptions import SniTunInvalidPeer
from .peer import Peer

_LOGGER = logging.getLogger(__name__)


class PeerManager:
    """Manage Peer connections."""

    def __init__(self, fernet_tokens: List[str]):
        """Initialize Peer Manager."""
        self._fernet = MultiFernet([Fernet(key) for key in fernet_tokens])
        self._peers = {}

    @property
    def connections(self):
        """Return count of connected devices."""
        return len(self._peers)

    def register_peer(self, fernet_data: bytes) -> Peer:
        """Create a new peer from crypt config."""
        try:
            data = self._fernet.decrypt(fernet_data).decode()
            config = json.loads(data)
        except (InvalidToken, json.JSONDecodeError):
            _LOGGER.warning("Invalid fernet token")
            raise SniTunInvalidPeer()

        # Check if token is valid
        if datetime.utcfromtimestamp(config['valid']) < datetime.utcnow():
            _LOGGER.info("Token was expired")
            raise SniTunInvalidPeer()

        # Extract configuration
        hostname = config['hostname']
        aes_key = bytes.fromhex(config['aes_key'])
        aes_iv = bytes.fromhex(config['aes_iv'])

        peer = self._peers[hostname] = Peer(hostname, aes_key, aes_iv)
        return peer

    def remove_peer(self, peer: Peer):
        """Remove peer from list."""
        self._peers.pop(peer.hostname, None)

    def peer_available(self, hostname: str) -> bool:
        """Check if peer available and return True or False."""
        if hostname in self._peers:
            return self._peers[hostname].is_ready
        return False

    def get_peer(self, hostname: str) -> Peer:
        """Get peer."""
        return self._peers.get(hostname)



"""Multiplexer channel."""
import asyncio
import logging
from ipaddress import IPv4Address
import uuid

import async_timeout

from ..utils.ipaddress import ip_address_to_bytes
from ..exceptions import MultiplexerTransportClose, MultiplexerTransportError
from .message import (CHANNEL_FLOW_CLOSE, CHANNEL_FLOW_DATA, CHANNEL_FLOW_NEW,
                      MultiplexerMessage)

_LOGGER = logging.getLogger(__name__)


class MultiplexerChannel:
    """Represent a multiplexer channel."""

    def __init__(self,
                 output: asyncio.Queue,
                 ip_address: IPv4Address,
                 channel_id=None) -> None:
        """Initialize Multiplexer Channel."""
        self._input = asyncio.Queue(5)
        self._output = output
        self._id = channel_id or uuid.uuid4()
        self._ip_address = ip_address

    @property
    def uuid(self) -> uuid.UUID:
        """Return UUID of this channel."""
        return self._id

    @property
    def ip_address(self):
        """Return caller IP4Address."""
        return self._ip_address

    async def write(self, data: bytes) -> None:
        """Send data to peer."""
        if not data:
            raise MultiplexerTransportError()

        # Create message
        message = MultiplexerMessage(self._id, CHANNEL_FLOW_DATA, data)

        try:
            async with async_timeout.timeout(5):
                await self._output.put(message)
        except asyncio.TimeoutError:
            _LOGGER.debug("Can't write to peer transport")
            raise MultiplexerTransportError() from None

    async def read(self) -> MultiplexerMessage:
        """Read data from peer."""
        message = await self._input.get()

        if message.flow_type == CHANNEL_FLOW_DATA:
            return message.data

        _LOGGER.debug("Read a close message for channel %s", self._id)
        raise MultiplexerTransportClose()

    def init_close(self) -> MultiplexerMessage:
        """Init close message for transport."""
        _LOGGER.debug("Close channel %s", self._id)
        return MultiplexerMessage(self._id, CHANNEL_FLOW_CLOSE)

    def init_new(self) -> MultiplexerMessage:
        """Init new session for transport."""
        _LOGGER.debug("New channel %s", self._id)
        extra = b"4" + ip_address_to_bytes(self.ip_address)
        return MultiplexerMessage(self._id, CHANNEL_FLOW_NEW, b"", extra)

    async def message_transport(self, message: MultiplexerMessage) -> None:
        """Only for internal ussage of core transport."""
        if self._input.full():
            _LOGGER.warning("Channel %s input is full", self._id)
            return

        await self._input.put(message)



"""Represent a single Peer."""
import asyncio
import hashlib
import logging
import os

from ..exceptions import MultiplexerTransportDecrypt, SniTunChallengeError
from ..multiplexer.core import Multiplexer
from ..multiplexer.crypto import CryptoTransport

_LOGGER = logging.getLogger(__name__)


class Peer:
    """Representation of a Peer."""

    def __init__(self, hostname: str, aes_key: bytes, aes_iv: bytes):
        """Initialize a Peer."""
        self._hostname = hostname
        self._multiplexer = None
        self._crypto = CryptoTransport(aes_key, aes_iv)

    @property
    def hostname(self) -> str:
        """Return his hostname."""
        return self._hostname

    @property
    def multiplexer(self) -> Multiplexer:
        """Return Multiplexer object."""
        return self._multiplexer

    @property
    def is_ready(self) -> bool:
        """Return true if the Peer is ready to process data."""
        if self._multiplexer is None:
            return False

        if self.multiplexer.wait().done():
            return False

        return True

    async def init_multiplexer_challenge(self, reader: asyncio.StreamReader,
                                         writer: asyncio.StreamWriter) -> None:
        """Initialize multiplexer."""
        try:
            token = hashlib.sha256(os.urandom(40)).digest()
            writer.write(self._crypto.encrypt(token))
            await writer.drain()

            data = await reader.readexactly(32)
            data = self._crypto.decrypt(data)

            # Check Token
            assert hashlib.sha256(token).digest() == data

        except (asyncio.IncompleteReadError, MultiplexerTransportDecrypt,
                AssertionError, OSError):
            _LOGGER.warning("Wrong challenge from peer")
            raise SniTunChallengeError()

        # Start Multiplexer
        self._multiplexer = Multiplexer(self._crypto, reader, writer)

    def wait_disconnect(self) -> None:
        """Wait until peer is disconnected.

        Return awaitable object.
        """
        if not self._multiplexer:
            raise RuntimeError("No Transport initialize for peer")

        return self._multiplexer.wait()



"""SniTun reference implementation."""
import asyncio
import logging
from typing import List

import async_timeout

from .listener_peer import PeerListener
from .listener_sni import SNIProxy
from .peer_manager import PeerManager

_LOGGER = logging.getLogger(__name__)


class SniTunServer:
    """SniTunServer helper class."""

    def __init__(self,
                 fernet_keys: List[str],
                 sni_port=None,
                 sni_host=None,
                 peer_port=None,
                 peer_host=None):
        """Initialize SniTun Server."""
        self._peers = PeerManager(fernet_keys)
        self._list_sni = SNIProxy(self._peers, host=sni_host, port=sni_port)
        self._list_peer = PeerListener(
            self._peers, host=peer_host, port=peer_port)

    @property
    def peers(self) -> PeerManager:
        """Return peer manager."""
        return self._peers

    def start(self):
        """Run server.

        Return coroutine.
        """
        return asyncio.wait([self._list_peer.start(), self._list_sni.start()])

    def stop(self):
        """Stop server.

        Return coroutine.
        """
        return asyncio.wait([self._list_peer.stop(), self._list_sni.stop()])


class SniTunServerSingle:
    """SniTunServer helper class."""

    def __init__(self, fernet_keys: List[str], host=None, port=None):
        """Initialize SniTun Server."""
        self._loop = asyncio.get_event_loop()
        self._peers = PeerManager(fernet_keys)
        self._list_sni = SNIProxy(self._peers, host=host)
        self._list_peer = PeerListener(self._peers, host=host)
        self._host = host
        self._port = port or 443
        self._server = None

    @property
    def peers(self) -> PeerManager:
        """Return peer manager."""
        return self._peers

    async def start(self):
        """Run server."""
        self._server = await asyncio.start_server(
            self._handler, host=self._host, port=self._port)

    async def stop(self):
        """Stop server."""
        self._server.close()
        await self._server.wait_closed()

    async def _handler(self, reader: asyncio.StreamReader,
                       writer: asyncio.StreamWriter):
        """Handle incoming connection."""
        try:
            async with async_timeout.timeout(2):
                data = await reader.read(2048)
        except asyncio.TimeoutError:
            _LOGGER.warning("Close handshake because timeout")
            writer.close()
            return

        # Connection closed / healty check
        if not data:
            writer.close()
            return

        # Select the correct handler for process data
        if data[0] == 0x16:
            self._loop.create_task(
                self._list_sni.handle_connection(reader, writer, data))
        else:
            self._loop.create_task(
                self._list_peer.handle_connection(reader, writer, data))



"""Connector to end resource."""
from contextlib import suppress
import asyncio
import logging
import ipaddress

from ..multiplexer.channel import MultiplexerChannel
from ..multiplexer.core import Multiplexer
from ..exceptions import MultiplexerTransportClose, MultiplexerTransportError

_LOGGER = logging.getLogger(__name__)


class Connector:
    """Connector to end resource."""

    def __init__(self, end_host: str, end_port=None, whitelist=False):
        """Initialize Connector."""
        self._loop = asyncio.get_event_loop()
        self._end_host = end_host
        self._end_port = end_port or 443
        self._whitelist = set()
        self._whitelist_enabled = whitelist

    @property
    def whitelist(self) -> set:
        """Allow to block requests per IP Return None or access to a set."""
        return self._whitelist

    def _whitelist_policy(self, ip_address: ipaddress.IPv4Address) -> bool:
        """Return True if the ip address can access to endpoint."""
        if self._whitelist_enabled:
            return ip_address in self._whitelist
        return True

    async def handler(self, multiplexer: Multiplexer,
                      channel: MultiplexerChannel) -> None:
        """Handle new connection from SNIProxy."""
        from_endpoint = None
        from_peer = None

        _LOGGER.debug("Receive from %s a request for %s", channel.ip_address,
                      self._end_host)

        # Check policy
        if not self._whitelist_policy(channel.ip_address):
            _LOGGER.warning("Block request from %s per policy",
                            channel.ip_address)
            await multiplexer.delete_channel(channel)
            return

        # Open connection to endpoint
        try:
            reader, writer = await asyncio.open_connection(
                host=self._end_host, port=self._end_port)
        except OSError:
            _LOGGER.error("Can't connect to endpoint %s:%s", self._end_host,
                          self._end_port)
            await multiplexer.delete_channel(channel)
            return

        try:
            # Process stream from multiplexer
            while not writer.transport.is_closing():
                if not from_endpoint:
                    from_endpoint = self._loop.create_task(reader.read(4096))
                if not from_peer:
                    from_peer = self._loop.create_task(channel.read())

                # Wait until data need to be processed
                await asyncio.wait([from_endpoint, from_peer],
                                   return_when=asyncio.FIRST_COMPLETED)

                # From proxy
                if from_endpoint.done():
                    if from_endpoint.exception():
                        raise from_endpoint.exception()

                    await channel.write(from_endpoint.result())
                    from_endpoint = None

                # From peer
                if from_peer.done():
                    if from_peer.exception():
                        raise from_peer.exception()

                    writer.write(from_peer.result())
                    from_peer = None

        except MultiplexerTransportError:
            _LOGGER.debug("Transport closed by endpoint for %s", channel.uuid)
            await multiplexer.delete_channel(channel)

        except MultiplexerTransportClose:
            _LOGGER.debug("Peer close connection for %s", channel.uuid)

        finally:
            if from_peer and not from_peer.done():
                from_peer.cancel()
            if from_endpoint and not from_endpoint.done():
                from_endpoint.cancel()
            with suppress(OSError):
                writer.close()
